/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *  ORBit: A CORBA v2.2 ORB
 *
 *  Copyright (C) 1998 Richard H. Porter
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Dick Porter <dick@cymru.net>
 *
 */

/*
 * Option parsing
 *
 * All ORB options are stripped from the application's argv, and argc is
 * adjusted accordingly
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "options.h"

void ORBit_option_parse(int *argc, char **argv, ORBit_orb_options *options)
{
	int i,j,numargs;
	char name[1024], *val;
	ORBit_orb_options *search=NULL;
	int *erase;
	ORBit_opt_type state=no_arg;

	numargs=*argc;

	erase=g_new0(int, *argc);

	for(i=1; i< *argc; i++) {
		/*
		 * if state!=no_arg, then search has been
		 * set.
		 */
		if(argv[i][0]!='-') {
			if(state==no_arg) {
				/* Skip non-option */
				continue;
			} else {
				/* an required option value has been found */
				erase[i]=1;
				numargs-=1;

				if(search->arg==NULL) {
					/* dont store any values, just strip
					 * the argv
					 */
					state=no_arg;
					continue;
				}

				if(state==string_arg) {
					char **str_arg=(char **)search->arg;

					/* free any existing value */
					if(*str_arg!=NULL) {
						g_free(*str_arg);
					}
					*str_arg=g_strdup(argv[i]);
				} else {
					/* this must be an int arg */
					int *int_arg=(int *)search->arg;

					*int_arg=atoi(argv[i]);
				}

				state=no_arg;
				continue;
			}
		} else {
			if(state==string_arg || state==int_arg) {
				fprintf(stderr, "Option %s requires an argument\n", search->name);
			}
		}

		val=argv[i];
		while(*val && *val=='-')
			val++;

		strncpy(name,val,1023);
		name[1023]='\0';

		val=strchr(name, '=');
		if(val!=NULL) {
			*val++='\0';
		}

		for(search=options;search->name!=NULL;search++) {
			if(!strcmp(name, search->name)) {
				break;
			}
		}

		if(search->name==NULL) {
			/* Didn't find it in our list of interesting options */
			state=no_arg;
		} else {
			/* Found it */
			erase[i]=1;
			numargs-=1;

			if(search->type==no_arg && search->arg!=NULL) {
				/* Treat as an int arg with val=1
				 */
				int *int_arg=(int *)search->arg;

				*int_arg=1;
			} else {
				if(val==NULL) {
					state=search->type;
				} else {
					if(search->type==string_arg &&
					   search->arg!=NULL) {
						char **str_arg=(char **)search->arg;

						/* free any existing value */
						if(*str_arg!=NULL) {
							g_free(*str_arg);
						}
						*str_arg=g_strdup(val);
					} else if(search->type==int_arg &&
						  search->arg!=NULL) {
						int *int_arg=(int *)search->arg;

						*int_arg=atoi(val);
					}

					state=no_arg;
				}
			}
		}
	}

	j=1;
	for(i=1; i< *argc; i++) {
		if(erase[i]==1) {
			continue;
		} else {
			if(j<numargs) {
				argv[j++]=argv[i];
			} else {
				argv[j++]='\0';
			}
		}
	}

	*argc=numargs;

	g_free(erase);
}
